@@ -221,6 +221,8 @@ class Config(AttrDict):
                 for cfg_name in values:
                     getattr(namespace, self.dest).rupdate(base_cfgs[cfg_name])
 
+                getattr(namespace, self.dest).resolve_transforms()
+
         class ConfigSetAction(argparse.Action):
             def __call__(
                 self,
@@ -263,16 +265,17 @@ class Config(AttrDict):
         group.add_argument(*unset_name, action=ConfigUnsetAction, nargs="+", dest=dest)
 
     @classmethod
-    def from_file(cls, filename: Union[str, Path]) -> Config:
+    def _load_file(cls, filename: Union[str, Path]) -> Config:
         filename = Path(filename)
-
         with filename.open() as f:
             cfg = Config(yaml.safe_load(f))
-
         cfg._resolve_cfg_id_after_file_loading()
-        cfg.resolve_base_cfgs()
-        cfg.resolve_transforms()
+        return cfg
 
+    @classmethod
+    def from_file(cls, filename: Union[str, Path]) -> Config:
+        cfg = cls._load_file(filename)
+        cfg.resolve_base_cfgs()
         return cfg
 
     @classmethod
@@ -289,13 +292,9 @@ class Config(AttrDict):
         cfg = Config()
 
         for k, fname in key_file_dict.items():
-            fname = Path(fname)
-            with fname.open() as f:
-                cfg[k] = Config(yaml.safe_load(f))
-                cfg[k]._resolve_cfg_id_after_file_loading()
+            cfg[k] = cls._load_file(fname)
 
         cfg.resolve_base_cfgs()
-        cfg.resolve_transforms()
 
         return cfg
 
@@ -305,6 +304,8 @@ class Config(AttrDict):
         for cfg_name in cfg_chain[1:]:
             ans.rupdate(self[cfg_name])
 
+        ans.resolve_transforms()
+
         if kv is None:
             return ans
 
@@ -590,7 +591,8 @@ class Config(AttrDict):
     ):
         if start_cfg is None:
             start_cfg = Config()
-        sdiff = start_cfg.diff(self)
+        start_cfg_transform_resolved = deepcopy(start_cfg).resolve_transforms()
+        sdiff = start_cfg_transform_resolved.diff(self)
         struct_diff, old_diff, new_diff = sdiff.diff_count()
         next_iter_cfgs = [
             (
@@ -603,7 +605,8 @@ class Config(AttrDict):
         for bname, bcfg in base_cfgs.items():
             cfg = deepcopy(start_cfg)
             cfg.rupdate(bcfg)
-            cdiff = cfg.diff(self)
+            cfg_transform_resolved = deepcopy(cfg).resolve_transforms()
+            cdiff = cfg_transform_resolved.diff(self)
             struct_diff, old_diff, new_diff = cdiff.diff_count()
             next_iter_cfgs.append(
                 (
@@ -645,7 +648,7 @@ class ConfigDiff(Config):
     class Nothing:
         pass
 
-    def __init__(self, cfg_old: Union[Config, Struct], cfg_new: Config):
+    def __init__(self, cfg_old: Union[Config, Struct], cfg_new: Union[Config, Struct]):
         super().__init__()
         self.set_attribute("_ConfigDiff__leaf", False)
 
@@ -668,6 +671,11 @@ class ConfigDiff(Config):
             self.set_new(cfg_new)
             return
 
+        assert isinstance(cfg_old, Config)
+        assert isinstance(cfg_new, Config)
+        cfg_old = deepcopy(cfg_old).finalize()
+        cfg_new = deepcopy(cfg_new).finalize()
+
         # create all_keys list in this manner to preserve order
         all_keys = list(cfg_old)
         for k in cfg_new:
